@@ -22,6 +22,7 @@ extern "C" {
 #endif
 
 extern enum pobj_arenas_assignment_type Default_arenas_assignment_type;
+extern size_t Default_arenas_max;
 
 #define HEAP_OFF_TO_PTR(heap, off) ((void *)((char *)((heap)->base) + (off)))
 #define HEAP_PTR_TO_OFF(heap, ptr)\
@@ -44,7 +45,8 @@ int heap_buckets_init(struct palloc_heap *heap);
 int heap_create_alloc_class_buckets(struct palloc_heap *heap,
 	struct alloc_class *c);
 
-int heap_extend(struct palloc_heap *heap, struct bucket *defb, size_t size);
+int heap_extend(struct palloc_heap *heap, struct bucket *defb,
+	size_t size);
 
 struct alloc_class *
 heap_get_best_class(struct palloc_heap *heap, size_t size);
@@ -54,13 +56,10 @@ heap_bucket_acquire(struct palloc_heap *heap, uint8_t class_id,
 		uint16_t arena_id);
 
 void
-heap_bucket_release(struct palloc_heap *heap, struct bucket *b);
+heap_bucket_release(struct bucket *b);
 
 int heap_get_bestfit_block(struct palloc_heap *heap, struct bucket *b,
 	struct memory_block *m);
-struct memory_block
-heap_coalesce_huge(struct palloc_heap *heap, struct bucket *b,
-	const struct memory_block *m);
 os_mutex_t *heap_get_run_lock(struct palloc_heap *heap,
 		uint32_t chunk_id);
 
@@ -96,7 +95,7 @@ unsigned heap_get_thread_arena_id(struct palloc_heap *heap);
 
 int heap_arena_create(struct palloc_heap *heap);
 
-struct bucket **
+struct bucket_locked **
 heap_get_arena_buckets(struct palloc_heap *heap, unsigned arena_id);
 
 int heap_get_arena_auto(struct palloc_heap *heap, unsigned arena_id);
@@ -106,6 +105,8 @@ int heap_set_arena_auto(struct palloc_heap *heap, unsigned arena_id,
 
 void heap_set_arena_thread(struct palloc_heap *heap, unsigned arena_id);
 
+unsigned heap_get_procs(void);
+
 void heap_vg_open(struct palloc_heap *heap, object_callback cb,
 		void *arg, int objects);
 
