@@ -1,34 +1,5 @@
-/*
- * Copyright 2015-2020, Intel Corporation
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- *     * Redistributions of source code must retain the above copyright
- *       notice, this list of conditions and the following disclaimer.
- *
- *     * Redistributions in binary form must reproduce the above copyright
- *       notice, this list of conditions and the following disclaimer in
- *       the documentation and/or other materials provided with the
- *       distribution.
- *
- *     * Neither the name of the copyright holder nor the names of its
- *       contributors may be used to endorse or promote products derived
- *       from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- */
+// SPDX-License-Identifier: BSD-3-Clause
+/* Copyright 2015-2021, Intel Corporation */
 
 /*
  * palloc.c -- implementation of pmalloc POSIX-like API
@@ -62,6 +33,7 @@
  *	3. memory blocks (sorted by lock address)
  */
 
+#include "bucket.h"
 #include "valgrind_internal.h"
 #include "heap_layout.h"
 #include "heap.h"
@@ -90,6 +62,7 @@ struct pobj_action_internal {
 		/* valid only when type == POBJ_ACTION_TYPE_HEAP */
 		struct {
 			uint64_t offset;
+			uint64_t usable_size;
 			enum memblock_state new_state;
 			struct memory_block m;
 			struct memory_block_reserved *mresv;
@@ -113,6 +86,9 @@ void
 palloc_set_value(struct palloc_heap *heap, struct pobj_action *act,
 	uint64_t *ptr, uint64_t value)
 {
+	/* suppress unused-parameter errors */
+	SUPPRESS_UNUSED(heap);
+
 	act->type = POBJ_ACTION_TYPE_MEM;
 
 	struct pobj_action_internal *actp = (struct pobj_action_internal *)act;
@@ -136,7 +112,7 @@ static int
 alloc_prep_block(struct palloc_heap *heap, const struct memory_block *m,
 	palloc_constr constructor, void *arg,
 	uint64_t extra_field, uint16_t object_flags,
-	uint64_t *offset_value)
+	struct pobj_action_internal *out)
 {
 	void *uptr = m->m_ops->get_user_data(m);
 	size_t usize = m->m_ops->get_user_size(m);
@@ -176,7 +152,8 @@ alloc_prep_block(struct palloc_heap *heap, const struct memory_block *m,
 	 * will be used to set the offset destination pointer provided by the
 	 * caller.
 	 */
-	*offset_value = HEAP_PTR_TO_OFF(heap, uptr);
+	out->offset = HEAP_PTR_TO_OFF(heap, uptr);
+	out->usable_size = usize;
 
 	return 0;
 }
@@ -247,7 +224,7 @@ palloc_reservation_create(struct palloc_heap *heap, size_t size,
 		goto out;
 
 	if (alloc_prep_block(heap, new_block, constructor, arg,
-		extra_field, object_flags, &out->offset) != 0) {
+		extra_field, object_flags, out) != 0) {
 		/*
 		 * Constructor returned non-zero value which means
 		 * the memory block reservation has to be rolled back.
@@ -265,14 +242,14 @@ palloc_reservation_create(struct palloc_heap *heap, size_t size,
 	 * The memory block cannot be put back into the global state unless
 	 * there are no active reservations.
 	 */
-	if ((out->mresv = b->active_memory_block) != NULL)
+	if ((out->mresv = bucket_active_block(b)) != NULL)
 		util_fetch_and_add64(&out->mresv->nresv, 1);
 
 	out->lock = new_block->m_ops->get_lock(new_block);
 	out->new_state = MEMBLOCK_ALLOCATED;
 
 out:
-	heap_bucket_release(heap, b);
+	heap_bucket_release(b);
 
 	if (err == 0)
 		return 0;
@@ -289,6 +266,9 @@ palloc_heap_action_exec(struct palloc_heap *heap,
 	const struct pobj_action_internal *act,
 	struct operation_context *ctx)
 {
+	/* suppress unused-parameter errors */
+	SUPPRESS_UNUSED(heap);
+
 #ifdef DEBUG
 	if (act->m.m_ops->get_state(&act->m) == act->new_state) {
 		ERR("invalid operation or heap corruption");
@@ -326,7 +306,7 @@ palloc_restore_free_chunk_state(struct palloc_heap *heap,
 				LOG(2, "unable to track runtime chunk state");
 			}
 		}
-		heap_bucket_release(heap, b);
+		heap_bucket_release(b);
 	}
 }
 
@@ -337,7 +317,8 @@ static void
 palloc_mem_action_noop(struct palloc_heap *heap,
 	struct pobj_action_internal *act)
 {
-
+	/* suppress unused-parameter errors */
+	SUPPRESS_UNUSED(heap, act);
 }
 
 /*
@@ -352,26 +333,18 @@ palloc_reservation_clear(struct palloc_heap *heap,
 		return;
 
 	struct memory_block_reserved *mresv = act->mresv;
-	struct bucket *b = mresv->bucket;
+	struct bucket_locked *locked = mresv->bucket;
 
 	if (!publish) {
-		util_mutex_lock(&b->lock);
-		struct memory_block *am = &b->active_memory_block->m;
-
 		/*
 		 * If a memory block used for the action is the currently active
-		 * memory block of the bucket it can be inserted back to the
+		 * memory block of the bucket it can be returned back to the
 		 * bucket. This way it will be available for future allocation
 		 * requests, improving performance.
 		 */
-		if (b->is_active &&
-		    am->chunk_id == act->m.chunk_id &&
-		    am->zone_id == act->m.zone_id) {
-			ASSERTeq(b->active_memory_block, mresv);
-			bucket_insert_block(b, &act->m);
-		}
-
-		util_mutex_unlock(&b->lock);
+		struct bucket *b = bucket_acquire(locked);
+		bucket_try_insert_attached_block(b, &act->m);
+		bucket_release(b);
 	}
 
 	if (util_fetch_and_sub64(&mresv->nresv, 1) == 1) {
@@ -423,12 +396,10 @@ palloc_heap_action_on_process(struct palloc_heap *heap,
 				act->m.m_ops->get_real_size(&act->m));
 		}
 	} else if (act->new_state == MEMBLOCK_FREE) {
-		if (On_valgrind) {
+		if (On_memcheck) {
 			void *ptr = act->m.m_ops->get_user_data(&act->m);
-			size_t size = act->m.m_ops->get_real_size(&act->m);
-
 			VALGRIND_DO_MEMPOOL_FREE(heap->layout, ptr);
-
+		} else if (On_pmemcheck) {
 			/*
 			 * The sync module, responsible for implementations of
 			 * persistent memory resident volatile variables,
@@ -441,6 +412,8 @@ palloc_heap_action_on_process(struct palloc_heap *heap,
 			 * that occur in newly allocated memory locations, that
 			 * once were occupied by a lock/volatile variable.
 			 */
+			void *ptr = act->m.m_ops->get_user_data(&act->m);
+			size_t size = act->m.m_ops->get_real_size(&act->m);
 			VALGRIND_REGISTER_PMEM_MAPPING(ptr, size);
 		}
 
@@ -477,6 +450,9 @@ palloc_mem_action_exec(struct palloc_heap *heap,
 	const struct pobj_action_internal *act,
 	struct operation_context *ctx)
 {
+	/* suppress unused-parameter errors */
+	SUPPRESS_UNUSED(heap);
+
 	operation_add_entry(ctx, act->ptr, act->value, ULOG_OPERATION_SET);
 }
 
@@ -555,8 +531,12 @@ palloc_exec_actions(struct palloc_heap *heap,
 	 * The operations array is sorted so that proper lock ordering is
 	 * ensured.
 	 */
-	qsort(actv, actvcnt, sizeof(struct pobj_action_internal),
-		palloc_action_compare);
+	if (actv) {
+		qsort(actv, actvcnt, sizeof(struct pobj_action_internal),
+			palloc_action_compare);
+	} else {
+		ASSERTeq(actvcnt, 0);
+	}
 
 	struct pobj_action_internal *act;
 	for (size_t i = 0; i < actvcnt; ++i) {
