@@ -1,34 +1,5 @@
-/*
- * Copyright 2015-2017, Intel Corporation
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- *     * Redistributions of source code must retain the above copyright
- *       notice, this list of conditions and the following disclaimer.
- *
- *     * Redistributions in binary form must reproduce the above copyright
- *       notice, this list of conditions and the following disclaimer in
- *       the documentation and/or other materials provided with the
- *       distribution.
- *
- *     * Neither the name of the copyright holder nor the names of its
- *       contributors may be used to endorse or promote products derived
- *       from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- */
+// SPDX-License-Identifier: BSD-3-Clause
+/* Copyright 2015-2020, Intel Corporation */
 
 /*
  * obj_basic_integration.c -- Basic integration tests
@@ -38,6 +9,7 @@
 #include <stddef.h>
 
 #include "unittest.h"
+#include "obj.h"
 
 #define TEST_STR "abcdefgh"
 #define TEST_STR_LEN 8
@@ -153,6 +125,9 @@ test_alloc_api(PMEMobjpool *pop)
 
 	POBJ_FREE(&node_zeroed);
 
+	PMEMoid oid = OID_NULL;
+	POBJ_FREE(&oid);
+
 	int err = 0;
 
 	err = pmemobj_alloc(pop, NULL, SIZE_MAX, 0, NULL, NULL);
@@ -221,6 +196,7 @@ test_realloc_api(PMEMobjpool *pop)
 
 	pmemobj_free(&oid);
 	UT_ASSERT(OID_IS_NULL(oid));
+	UT_ASSERTeq(pmemobj_alloc_usable_size(oid), 0);
 	UT_OUT("free");
 
 	/* alloc */
@@ -310,6 +286,11 @@ test_list_api(PMEMobjpool *pop)
 			sizeof(struct dummy_node), dummy_node_constructor,
 			&test_val);
 
+	TOID(struct dummy_node) inserted =
+			POBJ_LIST_FIRST(&D_RW(root)->dummies);
+	UT_ASSERTeq(pmemobj_type_num(inserted.oid),
+			TOID_TYPE_NUM(struct dummy_node));
+
 	TOID(struct dummy_node) node;
 	POBJ_ZNEW(pop, &node, struct dummy_node);
 
@@ -617,6 +598,27 @@ test_offsetof(void)
 				offsetof(struct dummy_node, plist_m));
 }
 
+static void
+test_layout(void)
+{
+	/* get number of declared types when there are no types declared */
+	POBJ_LAYOUT_BEGIN(mylayout);
+	POBJ_LAYOUT_END(mylayout);
+
+	size_t number_of_declared_types = POBJ_LAYOUT_TYPES_NUM(mylayout);
+	UT_ASSERTeq(number_of_declared_types, 0);
+}
+
+static void
+test_root_size(PMEMobjpool *pop)
+{
+	UT_ASSERTeq(pmemobj_root_size(pop), 0);
+
+	size_t alloc_size = sizeof(struct dummy_root);
+	pmemobj_root(pop, alloc_size);
+	UT_ASSERTeq(pmemobj_root_size(pop), sizeof(struct dummy_root));
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -625,26 +627,39 @@ main(int argc, char *argv[])
 	/* root doesn't count */
 	UT_COMPILE_ERROR_ON(POBJ_LAYOUT_TYPES_NUM(basic) != 2);
 
-	if (argc != 2)
-		UT_FATAL("usage: %s file-name", argv[0]);
+	if (argc < 2 || argc > 3)
+		UT_FATAL("usage: %s file-name [inject_fault]", argv[0]);
 
 	const char *path = argv[1];
-
+	const char *opt = argv[2];
 	PMEMobjpool *pop = NULL;
 
 	if ((pop = pmemobj_create(path, POBJ_LAYOUT_NAME(basic),
 			0, S_IWUSR | S_IRUSR)) == NULL)
 		UT_FATAL("!pmemobj_create: %s", path);
 
+	test_root_size(pop);
 	test_alloc_api(pop);
 	test_realloc_api(pop);
 	test_list_api(pop);
 	test_tx_api(pop);
 	test_action_api(pop);
 	test_offsetof();
+	test_layout();
 
 	pmemobj_close(pop);
 
+	/* fault injection */
+	if (argc == 3 && strcmp(opt, "inject_fault") == 0) {
+		if (pmemobj_fault_injection_enabled()) {
+			pmemobj_inject_fault_at(PMEM_MALLOC, 1,
+					"heap_check_remote");
+			pop = pmemobj_open(path, POBJ_LAYOUT_NAME(basic));
+			UT_ASSERTeq(pop, NULL);
+			UT_ASSERTeq(errno, ENOMEM);
+		}
+	}
+
 	if ((pop = pmemobj_open(path, POBJ_LAYOUT_NAME(basic))) == NULL)
 		UT_FATAL("!pmemobj_open: %s", path);
 
