@@ -1,34 +1,5 @@
-/*
- * Copyright 2015-2019, Intel Corporation
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- *     * Redistributions of source code must retain the above copyright
- *       notice, this list of conditions and the following disclaimer.
- *
- *     * Redistributions in binary form must reproduce the above copyright
- *       notice, this list of conditions and the following disclaimer in
- *       the documentation and/or other materials provided with the
- *       distribution.
- *
- *     * Neither the name of the copyright holder nor the names of its
- *       contributors may be used to endorse or promote products derived
- *       from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- */
+// SPDX-License-Identifier: BSD-3-Clause
+/* Copyright 2015-2020, Intel Corporation */
 
 /*
  * obj_tx_free.c -- unit test for pmemobj_tx_free
@@ -116,6 +87,121 @@ do_tx_free_wrong_uuid(PMEMobjpool *pop)
 	UT_ASSERT(!TOID_IS_NULL(obj));
 }
 
+/*
+ * do_tx_free_wrong_uuid_abort_on_failure -- try to free object with
+ * invalid uuid in a transaction where pmemobj_tx_set_failure_behavior
+ * was called.
+ */
+static void
+do_tx_free_wrong_uuid_abort_on_failure(PMEMobjpool *pop)
+{
+	volatile int ret = 0;
+	PMEMoid oid = do_tx_alloc(pop, TYPE_FREE_WRONG_UUID);
+	oid.pool_uuid_lo = ~oid.pool_uuid_lo;
+
+	/* pmemobj_tx_set_failure_behavior is called */
+	TX_BEGIN(pop) {
+		pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_RETURN);
+
+		UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+			POBJ_TX_FAILURE_RETURN);
+		ret = pmemobj_tx_free(oid);
+	} TX_ONCOMMIT {
+		UT_ASSERTeq(ret, EINVAL);
+	} TX_ONABORT {
+		UT_ASSERT(0); /* should not get to this point */
+	} TX_END
+
+	/* pmemobj_tx_set_failure_behavior is called */
+	TX_BEGIN(pop) {
+		pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_RETURN);
+
+		UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+			POBJ_TX_FAILURE_RETURN);
+		ret = pmemobj_tx_xfree(oid, 0);
+	} TX_ONCOMMIT {
+		UT_ASSERTeq(ret, EINVAL);
+	} TX_ONABORT {
+		UT_ASSERT(0); /* should not get to this point */
+	} TX_END
+
+	/* pmemobj_tx_set_failure_behavior is called in outer tx */
+	TX_BEGIN(pop) {
+		pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_RETURN);
+		TX_BEGIN(pop) {
+			UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+				POBJ_TX_FAILURE_RETURN);
+			ret = pmemobj_tx_free(oid);
+		} TX_ONCOMMIT {
+			UT_ASSERTeq(ret, EINVAL);
+		} TX_ONABORT {
+			UT_ASSERT(0); /* should not get to this point */
+		} TX_END
+		ret = pmemobj_tx_free(oid);
+	} TX_ONCOMMIT {
+		UT_ASSERTeq(ret, EINVAL);
+	} TX_ONABORT {
+		UT_ASSERT(0); /* should not get to this point */
+	} TX_END
+
+	/* pmemobj_tx_set_failure_behavior is called in neighbour tx */
+	TX_BEGIN(pop) {
+		TX_BEGIN(pop) {
+			pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_RETURN);
+			ret = pmemobj_tx_free(oid);
+		} TX_ONCOMMIT {
+			UT_ASSERTeq(ret, EINVAL);
+		} TX_ONABORT {
+			UT_ASSERT(0); /* should not get to this point */
+		} TX_END
+
+		TX_BEGIN(pop) {
+			UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+				POBJ_TX_FAILURE_ABORT);
+		} TX_ONCOMMIT {
+			UT_ASSERTeq(ret, EINVAL);
+		} TX_ONABORT {
+			UT_ASSERT(0); /* should not get to this point */
+		} TX_END
+	} TX_ONCOMMIT {
+		UT_ASSERTeq(ret, EINVAL);
+	} TX_ONABORT {
+		UT_ASSERT(0); /* should not get to this point */
+	} TX_END
+
+	/* pmemobj_tx_set_failure_behavior is called in neighbour tx */
+	TX_BEGIN(pop) {
+		pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_RETURN);
+		TX_BEGIN(pop) {
+			pmemobj_tx_set_failure_behavior(POBJ_TX_FAILURE_ABORT);
+			UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+				POBJ_TX_FAILURE_ABORT);
+		} TX_ONCOMMIT {
+			UT_ASSERTeq(ret, EINVAL);
+		} TX_ONABORT {
+			UT_ASSERT(0); /* should not get to this point */
+		} TX_END
+
+		TX_BEGIN(pop) {
+			UT_ASSERTeq(pmemobj_tx_get_failure_behavior(),
+				POBJ_TX_FAILURE_RETURN);
+			ret = pmemobj_tx_free(oid);
+		} TX_ONCOMMIT {
+			UT_ASSERTeq(ret, EINVAL);
+		} TX_ONABORT {
+			UT_ASSERT(0); /* should not get to this point */
+		} TX_END
+	} TX_ONCOMMIT {
+		UT_ASSERTeq(ret, EINVAL);
+	} TX_ONABORT {
+		UT_ASSERT(0); /* should not get to this point */
+	} TX_END
+
+	TOID(struct object) obj;
+	TOID_ASSIGN(obj, POBJ_FIRST_TYPE_NUM(pop, TYPE_FREE_WRONG_UUID));
+	UT_ASSERT(!TOID_IS_NULL(obj));
+}
+
 /*
  * do_tx_free_null_oid -- call pmemobj_tx_free with OID_NULL
  */
@@ -395,6 +481,8 @@ main(int argc, char *argv[])
 
 	do_tx_free_wrong_uuid(pop);
 	VALGRIND_WRITE_STATS;
+	do_tx_free_wrong_uuid_abort_on_failure(pop);
+	VALGRIND_WRITE_STATS;
 	do_tx_free_null_oid(pop);
 	VALGRIND_WRITE_STATS;
 	do_tx_free_commit(pop);
