@@ -1,6 +1,6 @@
 #!/usr/bin/env bash
 # SPDX-License-Identifier: BSD-3-Clause
-# Copyright 2016-2020, Intel Corporation
+# Copyright 2016-2022, Intel Corporation
 
 #
 # Used to check whether all the commit messages in a pull request
@@ -17,6 +17,7 @@ fi
 echo "Checking $1"
 
 subject=$(git log --format="%s" -n 1 $1)
+body=$(git log --format="%b" -n 1 $1)
 
 if [[ $subject =~ ^Merge.* ]]; then
 	# skip
@@ -28,6 +29,11 @@ if [[ $subject =~ ^Revert.* ]]; then
 	exit 0
 fi
 
+if [[ $body =~ "git-subtree-dir: src/deps/miniasync" ]]; then
+	# skip
+	exit 0
+fi
+
 # valid area names
 AREAS="pmem\|pmem2\|rpmem\|log\|blk\|obj\|pool\|set\|test\|benchmark\|examples\|doc\|core\|common\|daxio\|pmreorder"
 
